/* Copyright 1996, Jan D. Wolter and Steven R. Weiss, All Rights Reserved. */

/* Flat Textfile Identification Database Interface
 *
 * This module supports a user identification database formatted vaguely like
 * a standard unix /etc/passwd file.  Each line looks like:
 *
 *   <login>:<passwd>:<uid>:<gid>:<fname>:<dir>:<email>:<vc>:<priv>:<epriv>:
 *
 * Here <uid> is the user's uid number, <gid> is his primary group ID,
 * <fname> is the user's full name, and <dir> is the full path of the
 * directory where his configuration and participation files are saved.
 * <email> is his email address.  If the email address has been validated,
 * then <valcode> an empty string.  Otherwise, <vc> is the random
 * character code that must be given to validate the email address.
 * <priv> and <epriv> are integer privacy flags.
 */

#include "common.h"
#include "authident.h"

#include <ctype.h>
#include <errno.h>
#include <sys/stat.h>
#ifdef HAVE_SYS_FILE_H
#include <sys/file.h>
#endif

#include "set_uid.h"
#include "str.h"
#include "lock.h"
#include "entropy.h"

#ifdef DMALLOC
#include "dmalloc.h"
#endif

char *showopt_ident_module= "ident_text (file="PASSWD_FILE")";

/* FGETPW - Read a passwd-file-like line from the given file descriptor.
 * This vaguely like the Solaris fgetpwent() but stores in a different
 * structure.  THIS IS NOT VERY ROBUST!  It skips lines it doesn't understand.
 * It can cope with missing email and valcode fields, in which case it returns
 * those as NULL.
 */

struct btpasswd *fgetpw(FILE *fp)
{
    static char bf[BFSZ+1];
    static struct btpasswd pw;
    char *p, *b, *e;

    while ((p= fgets(bf, BFSZ, fp)) != NULL)
    {
	/* Skip lines longer than BFSZ */
	if ((p= strchr(bf,'\n')) == NULL)
	{
	    int ch;
	    while ((ch= fgetc(fp)) != '\n' && ch != EOF)
		;
	}
	else
	{
	    /* Get login name */
	    b= bf;
	    if ((e= strchr(b,':')) == NULL) continue;
	    pw.name= b; *e= '\0';

	    /* Get password */
	    b= e+1;
	    if ((e= strchr(b,':')) == NULL) continue;
	    pw.passwd= (e - b < 2) ? NULL : b; *e= '\0';

	    /* Get uid */
	    b= e+1;
	    if (!isascii(*b) || !isdigit(*b)) continue;
	    if ((e= strchr(++b,':')) == NULL) continue;
	    pw.uid= atoi(b);

	    /* Get gid */
	    b= e+1;
	    if ((e= strchr(b,':')) == NULL) continue;
	    pw.gid= atoi(b);
	    
	    /* Get fullname */
	    b= e+1;
	    if ((e= strchr(b,':')) == NULL) continue;
	    pw.gecos= b; *e= '\0';

	    pw.email= NULL;
	    pw.valcode= NULL;
	    pw.priv= -1;
	    pw.epriv= -1;
	    
	    /* Get directory */
	    b= e+1;
	    pw.dir= b;
	    if (*(e= firstin(b,":\n")) == '\0') return &pw;
	    *e= '\0';

	    /* Get email address */
	    b= e+1;
	    pw.email= b;
	    if (*(e= firstin(b,":\n")) == '\0') return &pw;
	    *e= '\0';

	    /* Get validation code */
	    b= e+1;
	    pw.valcode= b;
	    if (*(e= firstin(b,":\n")) == '\0') return &pw;
	    *e= '\0';

	    /* Get privacy code */
	    b= e+1;
	    if (*(e= firstin(b,":\n")) == '\0') return &pw;
	    if (isascii(*b) && isdigit(*b)) pw.priv= atoi(b);

	    /* Get email privacy code */
	    b= e+1;
	    if (*(e= firstin(b,":\n")) == '\0') return &pw;
	    if (isascii(*b) && isdigit(*b)) pw.epriv= atoi(b);

	    return &pw;
	}
    }
    return NULL;

}


/* GETPWNU - Get a particular name or uid from PASSWD_FILE.  Either the name
 * should be NULL or the uid should be -1, to indicate that the other is the
 * one we are searching for.  This is not a very fast implementation.
 */

struct btpasswd *getpwnu(char *nam, uid_t uid)
{
    FILE *pwfp;
    struct btpasswd *pw;

    set_httpd_if_shared();
    if ((pwfp= fopen(PASSWD_FILE,"r")) == NULL)
    {
	set_cfadm_if_shared();
	die("getpwnu cannot open user database "PASSWD_FILE);
    }
    lock_shared(fileno(pwfp),PASSWD_FILE);

    while ((pw= fgetpw(pwfp)) != NULL)
    {
	if ((uid >= 0 && pw->uid == uid) ||
	    (nam != NULL && !strcmp(nam, pw->name)))
		break;
    }

    fclose(pwfp);
    unlock(PASSWD_FILE);
    set_cfadm_if_shared();
    return pw;
}


/* MKPWLINE - Given a BFSZ buffer, and the fields of a btpasswd structure,
 * write a passwd text file line into the buffer.  Returns the length of
 * the line (not including trailing '\0').
 */

void mkpwline(char *bf, char *name, char *passwd, uid_t uid, gid_t gid,
	char *gecos, char *dir, char *email, char *valcode, int priv, int epriv)
{
    int rc;

#ifdef SHARE_AUTH_IDENT
    rc= snprintf(bf, BFSZ, "%s:%s:%d:%d:%.300s:%s:%.300s:%s:%d:%d:\n",
	    name, passwd, uid, gid, gecos, dir,
	    email ? email : "", valcode ? valcode : "", priv, epriv);
#else
    rc= snprintf(bf, BFSZ,"%s:*:%d:%d:%.300s:%s:%.300s:%s:%d:%d:\n",
	    name, uid, gid, gecos, dir,
	    email ? email : "", valcode ? valcode : "", priv, epriv);
#endif
    if (rc < 0 || rc > BFSZ -1)
	die("user fields too long for ident_text");
}


#ifdef HAVEIDENT_GETPWNAM
struct btpasswd *ident_getpwnam(char *login)
{
    return getpwnu(login,-1);
}
#endif


#ifdef HAVEIDENT_GETPWUID
struct btpasswd *ident_getpwuid(uid_t uid)
{
    return getpwnu(NULL,uid);
}
#endif


/* IDENT_WALKPW - This runs without locks, which may occasionally give glitches,
 * but this is never used for anything very critical (like authentication)
 * and we have no way of knowing when we are done with the file, so we'd
 * have to hold the lock till we exit.
 */

#ifdef HAVEIDENT_WALKPW
FILE *dbfp= NULL;
struct btpasswd *ident_walkpw(int flag)
{
    /* If flag is 0, goto begining of file */
    if (!flag)
    {
	if (dbfp == NULL)
	{
	    /* Open the file handle */
	    set_httpd_if_shared();
	    dbfp= fopen(PASSWD_FILE,"r");
	    set_cfadm_if_shared();
	    if (dbfp == NULL)
		die("resetpw cannot open user database "PASSWD_FILE);
	}
	else
	    /* Rewind the file handle */
	    rewind(dbfp);
    }
 
    return fgetpw(dbfp);
}
#endif


/* IDENT_NEWFNAME - change the user's fullname to the given value.
 */

#ifdef HAVEIDENT_NEWFNAME
void ident_newfname(struct btpasswd *pw, char *newfname)
{
    char bf[BFSZ];

    mkpwline(bf, pw->name, pw->passwd, pw->uid, pw->gid, newfname, pw->dir,
	pw->email, pw->valcode, pw->priv, pw->epriv);

    set_httpd_if_shared(); update_pw(PASSWD_FILE, pw->name, bf);
    set_cfadm_if_shared();

    flushusercache();
}
#endif


/* IDENT_NEWGID - change the user's groupid to the given value.
 */

#ifdef HAVEIDENT_NEWGID
void ident_newgid(struct btpasswd *pw, gid_t newgid)
{
    char bf[BFSZ];

    mkpwline(bf, pw->name, pw->passwd, pw->uid, newgid, pw->gecos, pw->dir,
	pw->email, pw->valcode, pw->priv, pw->epriv);

    set_httpd_if_shared();
    update_pw(PASSWD_FILE, pw->name, bf);
    set_cfadm_if_shared();

    flushusercache();
}
#endif


/* IDENT_NEWEMAIL - change the user's email address to the given value.
 */

#ifdef HAVEIDENT_NEWEMAIL
void ident_newemail(struct btpasswd *pw, char *newemail, int newvalid)
{
    char bf[BFSZ];
    char *newvalcode= (newemail[0] != '\0' && !newvalid) ? get_noise() : NULL;

    mkpwline(bf, pw->name, pw->passwd, pw->uid, pw->gid, pw->gecos, pw->dir,
	newemail, newvalcode, pw->priv, pw->epriv);

    set_httpd_if_shared();
    update_pw(PASSWD_FILE, pw->name, bf);
    set_cfadm_if_shared();

    flushusercache();
}
#endif


/* IDENT_NEWPRIV - change the user's privacy flags to the given value.
 */

#ifdef HAVEIDENT_NEWPRIV
void ident_newpriv(struct btpasswd *pw, int priv, int epriv)
{
    char bf[BFSZ];

    mkpwline(bf, pw->name, pw->passwd, pw->uid, pw->gid, pw->gecos, pw->dir,
	pw->email, pw->valcode, priv, epriv);

    set_httpd_if_shared();
    update_pw(PASSWD_FILE, pw->name, bf);
    set_cfadm_if_shared();

    flushusercache();
}
#endif

/* IDENT_ADD - Add an entry to the user database describing a user with
 * the given login, uid, gid, password, fullname, and directory.
 */

#ifdef HAVEIDENT_ADD
void ident_add(char *lid, uid_t uid, gid_t gid, char *fn, char *dir,
	char *email, int valid, int priv, int epriv)
{
    char bf[BFSZ];
    char *valcode= NULL;
    int fd;
    size_t len;

    if (email != NULL && email[0] != '\0' && !valid)
	valcode= get_noise();

    /* Construct line - note we never have shared auth/ident in this routine */
    len= mkpwline(bf, lid, NULL, uid, gid, fn, dir, email, valcode,
	    priv, epriv);

    if ((fd= open(PASSWD_FILE, O_WRONLY|O_APPEND)) < 0)
	die("ident_add cannot open "PASSWD_FILE" to write");
    lock_exclusive(fd,PASSWD_FILE);

    if (write(fd, bf, len) != len)
	die("write to "PASSWD_FILE" failed");
    
    close(fd);
    unlock(PASSWD_FILE);
}
#endif


/* DELIDENT - Delete the named user from the ident database.
 */

#ifdef HAVEIDENT_DEL
void ident_del(char *name)
{
    update_pw(PASSWD_FILE, name, NULL);
}
#endif
